#include "chess.h"
#include "data.h"
#include "weights.h"
//#include <math.h>

/*
    King Safety sub-terms
        Shield
        MAX(proximity, location)
        hits [number pieces attacking kings's 9'
		pawn 'storm'
        king has limited mobility {2d, linear only, none}
        
		no shield = 54-60
		MAX( center of board = 32+ ,	proximity = 16)
		R+M attack = 3+2 = 5
		pawn attacker = 10

		107/8 = 12
		107/4 = 25
		
*/
// q=12 r=5 m=3
//*
const int scale[128] ={
//                        R  2m      rm  3m  2r r2m q/4m 2rm r3m qm
//	  0   1   2   3   4   5   6   7   8   9  10  11  12  13  14  15
	  0,  0,  0,  0,  0,  0,  0,  0,  2,  4,  6,  8, 12, 14, 14, 16,
	 18, 20, 22, 24, 28, 30, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
//	 16  17  18  19  20  21  22  23  24  25  26  27  28  29  30  31
//	2R2m QR Q2m 2R3m QRm Q3mQ2R
	 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
	 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
	 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
	 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
	 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
	 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32};
//*/
/*
const int scale[128] ={//   2m      rm  3m q/2r r2m 4m  qm r3m  qr
	//0   1   2   3   4   5   6   7   8   9  10  11  12  13  14  15
	  0,  0,  0,  0,  0,  0,  0,  0,  2,  4,  8,  8, 12, 12, 14, 14,
	 16, 18, 18, 20, 22, 24, 25, 26, 27, 28, 29, 30, 31, 32, 32, 32,
//	 16  17 QRm  19  20  21  22  23  24  25  26  27  28  29  30  31
	 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
	 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
	 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
	 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
	 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
	 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32};
*/

void cBOARD::eval_king_safety(){
    int sq;
    int r, f, s, la, pa, i, d;//, h;//, es, se;
    unsigned int attack;
    
    sq = pieces[WHITE][KING];
    
	// q=16 p=8 q/2p + help += 16
	d = proximity[BLACK];
	d += (int)(d>24) << 3; // << piece_count[BLACK][QUEEN]
	s = MAX(king_defects[WHITE][sq],d-(defenders[WHITE]));
    
    attack = 0;
    i = 0;
	r = 1;
    do{
        d = sq + delta[i];
        //attack += ((d & OFFBOARD) || (control[d] & 0xffff0000) > 0 || square[d].color == WHITE);
        if((d & OFFBOARD) || (control[d] & 0xffff0000) > 0 || square[d].color == WHITE);
		else attack |= r;
		if(i == 7) break;
        i++;
		r <<= 1;
    }while(true);

	// lat  diag  `\ clockwise 
	// 8765 4321  <- clockwise
	// how to tell king is NOT confined to 1 line of freedom
	// or HAS more than 1 degree of freedom
	// 1 dig line == 0101 or 1010
	// 1 lat line == 0101 or 1010 so as long as any diag AND any lat OR any even AND any odd
	//if(((attack&240) && (attack&15)) || ((attack&85) && (attack&170)))
    
    la = 0;
    pa = 0;
    r = sq >> 4;
    f = (sq & 7) + 1;
    if(f==1){ f = 2; sq++; } // pro: doesnt move into corner
    if(f==8){ f = 7; sq--; } // con: 
    
    f--;
    if(least_adv_pawn[WHITE][f] == 0){  // white pawn is gone 
        //la += lat_onfile[BLACK][f] * ropenbyking; // open file
        if(least_adv_pawn[BLACK][f] == 7) s += shield5; // black pawn is gone too
        else s += shield4;
    }else{ // white pawn is here somewhere...
        //la += lat_onfile[BLACK][f] * rhopenbyking; // half open file
		if(r > 3){
			if(square[sq-17].iswhitepawn());// no penalty
			else if(square[sq-1].iswhitepawn()) s += shield0;
			else if(square[sq-33].iswhitepawn()) s += shield1;
			else if(square[sq-49].iswhitepawn()) s += shield2;
			else s += shield3;
		}else s += shield4;
    }
    if(most_adv_pawn[BLACK][f] > 0 && most_adv_pawn[BLACK][f] <= r) // avoid negative values
        pa += pawn_aggressor[r-most_adv_pawn[BLACK][f]];
    
    f++;
    if(least_adv_pawn[WHITE][f] == 0){
        //la += lat_onfile[BLACK][f] * ropenbyking; // open file
        if(least_adv_pawn[BLACK][f] == 7) s += shield6;
        else s += shield5;
    }else{
        //la += lat_onfile[BLACK][f] * rhopenbyking; // half open file
        if(r > 3){
			if(square[sq-16].iswhitepawn());// no penalty
			else if(square[sq-0].iswhitepawn()) s += shield0;
			else if(square[sq-32].iswhitepawn()) s += shield1;
			else if(square[sq-48].iswhitepawn()) s += shield2;
			else s += shield3;
		}else s += shield5;
    }
    if(most_adv_pawn[BLACK][f] > 0 && most_adv_pawn[BLACK][f] <= r) // avoid negative values
        pa += pawn_aggressor[r-most_adv_pawn[BLACK][f]];
    
    f++;
    if(least_adv_pawn[WHITE][f] == 0){
        //la += lat_onfile[BLACK][f] * ropenbyking; // open file
        if(least_adv_pawn[BLACK][f] == 7) s += shield5;
        else s += shield4;
    }else{
        //la += lat_onfile[BLACK][f] * rhopenbyking; // half open file
        if(r > 3){
			if(square[sq-15].iswhitepawn());// no penalty
			else if(square[sq+1].iswhitepawn()) s += shield0;
			else if(square[sq-31].iswhitepawn()) s += shield1;
			else if(square[sq-47].iswhitepawn()) s += shield2;
			else s += shield3;
		}else s += shield4;
    }
    if(most_adv_pawn[BLACK][f] > 0 && most_adv_pawn[BLACK][f] <= r) // avoid negative values
        pa += pawn_aggressor[r-most_adv_pawn[BLACK][f]];
    
    //s += (attack == 8) * no_escape;
	//cout << attack << endl;
	if(attack == 0) s += no_escape;
	else if(((attack&240) && (attack&15)) || ((attack&85) && (attack&170)));
	else{
		s += no_escape * 3 / 4;
		//cout << "king limited to one degree of freedom";
	}
    // Q=3 R=2 m=1
	// 0, 1, 2, 3, 4, 5, 6, 7, 8
	// 0  2  8 10 12 14 16 18
	// 0  4 16 20 24 28 32 36
	d = hits[BLACK];
	d *= d; //+= (int)(d>1) << 1;
	
	s += pa + d;//(d << 1); 
	
	s = s * scale[force[BLACK]] / 32;
	

	s = king_safe[s];
    
	liability[WHITE] += s;

    
    
    sq = pieces[BLACK][KING];
	d = proximity[WHITE];
	d += (int)(d>24) << 3; // << piece_count[WHITE][QUEEN]
	s = MAX(king_defects[BLACK][sq],d-(defenders[BLACK]));
    
    attack = 0;
    i = 0;
	r = 1;
    do{
        d = sq + delta[i];
        //attack += ((d & OFFBOARD) || (control[d] & 0x0000ffff) > 0 || square[d].color == BLACK);
        if((d & OFFBOARD) || (control[d] & 0x0000ffff) > 0 || square[d].color == BLACK);
		else attack |= r;
		if(i == 7) break;
        i++;
		r <<= 1;
    }while(true);
    
    
    la = 0;
    pa = 0;
    r = sq >> 4;
    f = (sq & 7) + 1;
    if(f==1){ f = 2; sq++; } // pro: doesnt move into corner
    if(f==8){ f = 7; sq--; } // con: 
    
    f--;
    if(least_adv_pawn[BLACK][f] == 7){
        //la += lat_onfile[WHITE][f] * ropenbyking; // open file
        if(least_adv_pawn[WHITE][f] == 0) s += shield5;
        else s += shield4;
    }else{
        //la += lat_onfile[WHITE][f] * rhopenbyking; // half open file
		if(r<4){
			if(square[sq+15].isblackpawn()); // no penalty
			else if(square[sq-1].isblackpawn()) s += shield0;
			else if(square[sq+31].isblackpawn()) s += shield1;
			else if(square[sq+47].isblackpawn()) s += shield2;
			else s += shield3;
		}else s += shield4;
    }
    if(most_adv_pawn[WHITE][f] < 7 && most_adv_pawn[WHITE][f] >= r) // avoid negative values
        pa += pawn_aggressor[most_adv_pawn[WHITE][f]-r];
    
    f++;
    if(least_adv_pawn[BLACK][f] == 7){
        //la += lat_onfile[WHITE][f] * ropenbyking; // open file
        if(least_adv_pawn[WHITE][f] == 0) s += shield6;
        else s += shield5;
    }else{
        //la += lat_onfile[WHITE][f] * rhopenbyking; // half open file
        if(r<4){
			if(square[sq+16].isblackpawn()); // no penalty
			else if(square[sq-0].isblackpawn()) s += shield0;
			else if(square[sq+32].isblackpawn()) s += shield1;
			else if(square[sq+48].isblackpawn()) s += shield2;
			else s += shield3;
		}else s += shield5;
    }
    if(most_adv_pawn[WHITE][f] < 7 && most_adv_pawn[WHITE][f] >= r) // avoid negative values
        pa += pawn_aggressor[most_adv_pawn[WHITE][f]-r];
    
    f++;
    if(least_adv_pawn[BLACK][f] == 7){
        //la += lat_onfile[WHITE][f] * ropenbyking; // open file
        if(least_adv_pawn[WHITE][f] == 0) s += shield5;
        else s += shield4;
    }else{
        //la += lat_onfile[WHITE][f] * rhopenbyking; // half open file
        if(r<4){
			if(square[sq+17].isblackpawn()); // no penalty
			else if(square[sq+1].isblackpawn()) s += shield0;
			else if(square[sq+33].isblackpawn()) s += shield1;
			else if(square[sq+49].isblackpawn()) s += shield2;
			else s += shield3;
		}else s += shield4;
    }
    if(most_adv_pawn[WHITE][f] < 7 && most_adv_pawn[WHITE][f] >= r) // avoid negative values
        pa += pawn_aggressor[most_adv_pawn[WHITE][f]-r];
    
    //cout << "shield = " << s << endl;
	
	//s += (attack == 8) * no_escape;
	//cout << attack << endl;
	if(attack == 0) s += no_escape;
	else if(((attack&240) && (attack&15)) || ((attack&85) && (attack&170)));
	else{
		s += no_escape * 3 / 4;
		//cout << "king limited to one degree of freedom";
	}
    
	//cout << "+= no escape = " << s << endl;

	//s -= (s * undeveloped[WHITE]) >> 3; // nothing developed = -50%  19 points for 1
    
	d = hits[WHITE];
	d *= d;//+= (int)(d>1) << 1;
	
	s += pa + d;//(d << 1); 
	
	s = s * scale[force[WHITE]] / 32;
	
	/*if(s > qdepth){  *** basically < 150 ***
		qdepth = s;
		cout << s << " ";
	}*/
	
	s = king_safe[s];

	liability[BLACK] += s;
	
}
